/*
 * libkarma/properties.c
 * 
 * Copyright (c) Frank Zschockelt <libkarma@freakysoft.de> 2004
 * 
 * You may distribute and modify this program under the terms of 
 * the GNU GPL, version 2 or later.
 * 
 */
#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <errno.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>
#include <utime.h>
#include <iconv.h>

#include "properties.h"
#include "util.h"
#include "errors.h"


struct klist props = {NULL, 0, 256, (iconv_t)(-1), (iconv_t)(-1), NULL};
/* props.in, props.out need to be initialized for change_codeset proper work */


void init_properties(char * codeset)
{
	props.l=NULL;
	props.count=0;
	props.highest_fid=256;
	props.charset=strdup(codeset);
	change_codeset(codeset);
}

void destroy_properties(void)
{
	struct list * tmp;

	while(props.l != NULL) {
		tmp=props.l;
		props.l=tmp->next;

		hash_destroy(&(tmp->element));
		free(tmp);
	}

	props.count=0;
	props.highest_fid=256;
	free(props.charset);

	if(iconv_close(props.in) || iconv_close(props.out))
		lkerrSet(E_ICONVC);
	props.in=props.out=(iconv_t)-1;
}

char * current_codeset(void)
{
	return props.charset;
}

void change_codeset(char *codeset)
{
	int errIn=0, errOut=0;
	if (props.in  != (iconv_t)(-1)) {
		errIn  = iconv_close(props.in);  
		props.in  = (iconv_t)(-1);
	}
	if (props.out != (iconv_t)(-1)) {
		errOut = iconv_close(props.out); 
		props.out = (iconv_t)(-1);
	}
	if (errIn || errOut) {
		lkerrSet(E_ICONVC);
		return;
	}

	props.in  = iconv_open("UTF-8", codeset);
	props.out = iconv_open(codeset, "UTF-8");
	if ((props.in  == (iconv_t)-1) || (props.out == (iconv_t)-1)) {
		lkerrSet(E_ICONVO);
	}
	free(props.charset);
	props.charset=strdup(codeset);
}

uint32_t import_properties(char * properties)
{
	uint32_t count=0;
	char *key, *value;
	char *ikey, *ivalue;
	struct list *item, *l; /* unused: *o; */
	char *n_tok, *last;

	last=&(properties[strlen(properties)])+1; 
	/* otherwise we can't see whether n_tok is outside of properties */

	item=malloc(sizeof(struct list));
	item->fid=0;
	item->element=hash_create(32); /* 32 should be nice*/
	key=strtok_r(properties, "=", &n_tok);
	if(key==NULL){
		lkerrSet(E_BADPROP);
		return -1;
	}						
	value=strtok_r(NULL, "\n", &n_tok);

	while(value != NULL){
		if(strcmp(key, "fid") == 0)
			item->fid=strtoul(value, (char **)NULL, 10);

		ikey=strdup(key);
		ivalue=strdup(value);
		hash_insert(item->element, ikey, ivalue);

		if((n_tok[0] == '\n')||(n_tok[0]=='\0')){
			++n_tok;

			if((props.l == NULL) || (props.l->fid > item->fid)){
				item->next=props.l;
				props.l=item;
			}else{
				l=props.l;
				while((l->next != NULL) && 
						(l->next->fid <= item->fid))
					l=l->next;

				if(l->fid == item->fid){
					hash_destroy(&l->element);
					l->element=item->element;
					free(item);
					item=NULL;
				}else{
					item->next=l->next;
					l->next=item;
				}
			}

			if(item){
				if(item->fid > props.highest_fid)
					props.highest_fid=item->fid;
				++props.count;
			}

			item=malloc(sizeof(struct list));
			item->element=hash_create(32);
			++count;
		}
		if(n_tok==last)
			key=value=NULL;
		else{
			key=strtok_r(NULL, "=", &n_tok);
			value=strtok_r(NULL, "\n", &n_tok);
		}
	}
	hash_destroy(&item->element);
	free(item);
	return count;
}

char * export_property(uint32_t id)
{
	HASH *table; struct hash_elem *tmp; 
	char *p;
	int len, i;

	/* 	void export_hash(const char *key, void *data) */
#define	export_hash(key, data) { \
	len+=strlen(key)+strlen((char *)data)+6; \
	p=realloc(p, len); \
	strcat(p, key); \
	strcat(p, "="); \
	strcat(p, (char*)data); \
	strcat(p, "\n"); \
	}

	len=0;
	p=malloc(1);
	*p=0;
	table=idsearch(id);
	if (table==NULL) {
		lkerrSet(E_BADID);
		return NULL;
	}
	else {  /* hash_walker(table, export_hash); */
		for (i=0; i < table->nelem; i++) {
			tmp=table->table[i];
			while(tmp){
				export_hash(tmp->key, tmp->data);
				tmp=tmp->next;
			}
		}
	}
	return p;
}

/* The karma only plays files when the ID is > 256 and ID%16==0 */
uint32_t getnextfid(void)
{
	struct list *t;
	uint32_t smallest_fid=272;
	t=props.l;
	while((t!=NULL) && (t->fid<=smallest_fid)){
		if(t->fid==smallest_fid)
			smallest_fid+=16;
		t=t->next;
	}
	return smallest_fid;
}

uint32_t new_property(void)
{
	struct list *item, *l;

	item=malloc(sizeof(struct list));
	item->element=hash_create(32);
	item->next=NULL;
	item->fid=getnextfid();
	change_property_hash(item->element, "fid", simple_itoa(item->fid));

	if((props.l == NULL) || (props.l->fid > item->fid)){
		item->next=props.l;
		props.l=item;
	}else{
		l=props.l;
		while((l->next != NULL)&&(l->next->fid < item->fid))
			l=l->next;
		item->next=l->next;
		l->next=item;
	}
	if(item->fid > props.highest_fid)
		props.highest_fid=item->fid;
	props.count++;
	return item->fid;
}


char * get_property(uint32_t fid, char * key)
{
	HASH *tmp;

	tmp=idsearch(fid);
	if(tmp==NULL){
		lkerrSet(E_BADID);
		return NULL;
	}
	else return (char *)hash_find(tmp, key);
}

char * get_propertyUtf8(uint32_t fid, char * key)
{
	HASH *tmp;
	char *aux;

	tmp=idsearch(fid);
	if(tmp==NULL){
		lkerrSet(E_BADID);
		return NULL;
	}
	aux=(char *)hash_find(tmp, key);
	return convertFromUtf8(aux);
}

char * get_property_hash(HASH * p, char * key)
{
	return (char *)hash_find(p, key);
}

int delete_property(uint32_t id)
{
	struct list *l,*p;

	l=props.l;
	p=props.l;

	if(p==NULL){
		lkerrSet(E_BADID);
		return -1;
	}

	while(p->fid != id){
		l=p;
		p=p->next;
	}

	if(props.l==l)
		props.l=p->next;
	else l->next=p->next;

	if(p->fid==props.highest_fid){
		if(l!=NULL)
			props.highest_fid=l->fid;
		else
			props.highest_fid=256;
	}

	hash_destroy(&(p->element));
	free(p);
	props.count--;
	return 0;
}

char * convert2utf8(char * data)
{
	char *ivalue, *auxdata, *auxvalue;
	size_t insz, outsz, nconv;

	if (data == NULL) return NULL;

	auxdata = data;
	insz = strlen(auxdata);
	outsz = 2 * insz;
	ivalue = (char *)malloc(outsz+1);
	ivalue[0] = '\0';
	auxvalue = ivalue;
	nconv = iconv(props.in, &auxdata, &insz, &auxvalue, &outsz);
	auxdata = data;  /* *auxdata has been modified by iconv ... */
	if (nconv == (size_t)-1) {
		lkerrSet(E_ICONV);
		ivalue=strdup(data);
	}
	else auxvalue[0] = '\0';
	ivalue = realloc(ivalue, strlen(ivalue)+1);

	return ivalue;
}

char * convertFromUtf8(char * data) {
	char *ivalue, *auxdata, *auxvalue;
	size_t insz, outsz, nconv;

	if (data == NULL) return NULL;
	
	auxdata=data;
	insz = strlen(auxdata);
	outsz = 2 * insz;
	ivalue = (char *)malloc(outsz+1);
	ivalue[0] = '\0';
	auxvalue = ivalue;
	nconv = iconv(props.out, &auxdata, &insz, &auxvalue, &outsz);
	auxdata = data;  /* *auxdata has been modified by iconv ... */
	if (nconv == (size_t)-1) {
 		lkerrSet(E_ICONV);
		ivalue=strdup(data);
	}
	else auxvalue[0] = '\0';
	ivalue = realloc(ivalue, strlen(ivalue)+1);

	return ivalue;
}

int change_property(uint32_t id, char * property, char * data)
{
	HASH * p;

	p=idsearch(id);
	if(p==NULL){
		lkerrSet(E_BADID);
		return -1;
	}
	change_property_hash(p, property, data);

	return 0;
}

void change_property_hash(HASH * p, char * property, char * data)
{
	char *ikey, *ivalue;

	if (data == NULL) return;

	ikey = strdup(property);
	ivalue = convert2utf8(data);
	hash_insert(p, ikey, ivalue);
}

void change_property_hash_raw(HASH * p, char * property, char * data)
{
	char *ikey, *ivalue;

	if (data == NULL) return;

	ikey = strdup(property);
	ivalue = strdup(data);
	hash_insert(p, ikey, ivalue);
}

/* ------------------------------------------------------------------- TO DELETE
void change_property_hash(HASH * p, char * property, char * data)
{
        char *ikey, *ivalue, *auxvalue, *auxdata;
	size_t insz, outsz, nconv;

	if (data != NULL) {
		ikey = strdup(property);
		auxdata = data;
		insz = strlen(auxdata);
		outsz = 2 * insz;
		ivalue = (char *)malloc(outsz+1);
		ivalue[0] = '\0';
		auxvalue = ivalue;
		nconv = iconv(props.in, &auxdata, &insz, &auxvalue, &outsz);
		auxdata = data;  // *auxdata has been modified by iconv ... //
		if (nconv == (size_t)-1) {
			lkerrSet(E_ICONV);
			ivalue=strdup(data);
		}
		else auxvalue[0] = '\0';
		ivalue = realloc(ivalue, strlen(ivalue)+1);
		hash_insert(p, ikey, ivalue);
	}
}
--------------------------------------------------------------- END TO DELETE */

HASH * idsearch(uint32_t fid)
{
	struct list * ptr;

	ptr=props.l;
	while((ptr!=NULL)&&(ptr->fid < fid))
		ptr=ptr->next;
	if((ptr == NULL) || (ptr->fid != fid)) return NULL;
	else return ptr->element;
}

int load_properties(void)
{
	int fd;
	char * buf;
	struct stat info;
	char *serial;
	char *fname;

	fname=strdup(getenv("HOME"));
	/*33=strlen("/.openrio/pearl-0123456/fileinfo")*/
	fname=realloc(fname, strlen(fname)+33);
	serial = (char *)calloc(8, sizeof(char));
	snprintf(serial, 7, "%d", atoi(get_property(0,"serial")));
	strcat(fname,"/.openrio/pearl-");
	strcat(fname,serial);
	strcat(fname,"/fileinfo");

	fd=open(fname,O_RDONLY);
	if(fd==-1) {
		free(fname);
		free(serial);
		lkerrSet(E_OPEN);
		return -1;
	}
	fstat(fd, &info);
	buf=(char *)malloc((int)info.st_size+1);
	read(fd, buf, info.st_size);
	close(fd);
	buf[info.st_size]='\0';
	import_properties(buf);
	free(buf);
	free(fname);
	free(serial);
	return 0;
}
/*
   caches are stored in $HOME/.openrio/pearl-$SERIAL.

   The $SERIAL value is the Karma's unique serial number.
   This allows multiple Karmas to sync against the same machine without 
   collisions.

   File Format
   At this time, there is only one file specified within the cache 
   directory, named fileinfo. 
   This is a standard text file containing a dump of the File Info properties, 
   as returned by the Get All File Info message to the device.

   The file's last-modified date is set to the device_generation value provided 
   by the Karma  essentially, the last-modified date of the Karma's internal 
   database. If the cache file's last-modified date and the Karma's 
   device_generation do not match, the cache is considered invalid and is 
   rebuilt using information from the device. 
   */
int save_properties(void)
{
	int fd;
	uint32_t i;
	char * serial;
	char * fname;
	char * buf;
	struct list *tmp;
	struct utimbuf time;

	fname=strdup(getenv("HOME"));
	/*33=strlen("/.openrio/pearl-0123456/fileinfo")*/
	fname=realloc(fname, strlen(fname)+33);
	serial = (char *)calloc(8, sizeof(char));
	snprintf(serial, 7, "%d", atoi(get_property(0,"serial")));
	strcat(fname,"/.openrio/");
	strcat(fname,"pearl-");
	strcat(fname,serial);
	strcat(fname, "/");
	strcat(fname,"fileinfo");
	if(create_path(fname) == -1){
		free(fname);
		free(serial);
		return -1;
	}
	fd=open(fname,O_CREAT|O_WRONLY|O_TRUNC, S_IRUSR|S_IWUSR);
	if(fd==-1) {
		free(fname);
		free(serial);
		lkerrSet(E_OPEN);
		return -1;
	}

	tmp=props.l;
	if(tmp->next) tmp=tmp->next; /* don't save device_settings (fid==0)*/
	for(i=0; i<props.count-1; i++) {
		buf=export_property(tmp->fid);
		write(fd, buf, strlen(buf));
		free(buf);
		write(fd, "\n", 1);
		tmp=tmp->next;
	}

	close(fd);

	time.actime=strtoul(get_property(0, "device_generation"), NULL, 10);
	time.modtime=time.actime;
	if(utime(fname, &time)==-1){
		free(fname);
		free(serial);
		lkerrSet(E_UTIME);
		return -1;
	}

	free(fname);
	free(serial);
	return 0;
}

void inc_devicegeneration(void)
{
	uint32_t dg;

	dg=strtoul(get_property(0, "device_generation"), (char **)NULL, 10);
	change_property(0, "device_generation", simple_itoa(dg+1));
}


int cache_obsolete(char *device_settings)
{
   	char *serial, *fname, *tmp;
	struct stat modified;

	tmp=strdup(device_settings);
	import_properties(tmp);
	free(tmp);

	fname=strdup(getenv("HOME"));
	/*33=strlen("/.openrio/pearl-0123456/fileinfo")*/
	fname=realloc(fname, strlen(fname)+33);
	serial = (char *)calloc(8, sizeof(char));
	snprintf(serial, 7, "%d", atoi(get_property(0,"serial")));
	strcat(fname,"/.openrio/pearl-");
	strcat(fname,serial);
	strcat(fname,"/fileinfo");
	if(stat(fname, &modified) == -1){
		free(fname);
		free(serial);
		return 1;
	}
	free(fname);
	free(serial);
	if(strtoul(get_property(0, "device_generation"), NULL, 10) == 
			(unsigned)modified.st_mtime){
		return 0;
	}else return 1;
}

int db_changed(char *device_settings)
{
	char *old, *tmp;
	int ret;

	tmp=get_property(0, "device_generation");
	if(tmp){
		old=strdup(tmp);
		tmp=strdup(device_settings);
		import_properties(tmp);
		free(tmp);
		ret=strcmp(get_property(0, "device_generation"), old) != 0;
		free(old);
		return ret;
	}else return 1;
}

uint32_t *andOrSearch(int andOr, uint32_t *search_in, char *key, char *data)
{
	struct list *ptr;
	char * val, *ival;
	int hash;
	int i,j;
	uint32_t * result;
	
	result=malloc(sizeof(uint32_t));
	result[0]=0;
	i=1;

	ival = convert2utf8(data);

	if(search_in==NULL){
		ptr=props.l;
		hash=hash_string((const unsigned char *)key, 
				ptr->element->nelem);
		while(ptr->next!=NULL){
			ptr=ptr->next;
			val=(char *)hash_find_hash(ptr->element, hash, key);
			if((val!=NULL) && (strstr(val, ival) != NULL)){
				result[i-1]=ptr->fid;
				++i;
				result=realloc(result, i*sizeof(uint32_t));
				result[i-1]=0;
			}
		}
	}
	else if (andOr) {                                        /* AND merge */
		j=0;
		while(search_in[j] != 0){
			val=get_property(search_in[j] , key);
			if((val!=NULL) && (strstr(val, ival) != NULL)){
			  	result[i-1]=search_in[j];
			  	++i;
			  	result=realloc(result, i*sizeof(uint32_t));
			  	result[i-1]=0;
			}
			++j;
		}
		free(search_in);
	}else{                                                    /* OR merge */
		j=0;
		while(search_in[j] != 0){
			result[i-1]=search_in[j];
			++i;
			result=realloc(result, i*sizeof(uint32_t));
			result[i-1]=0;
			++j;
		}
		ptr=props.l;
		hash=hash_string((const unsigned char *)key, 
				ptr->element->nelem);
		while(ptr->next!=NULL){
		  ptr=ptr->next;
		  val=(char *)hash_find_hash(ptr->element, hash, key);
		  if((val!=NULL) && (strstr(val, ival) != NULL)){
		    j=0;
		    while((result[j] != ptr->fid) && (result[j] != 0))
		      j++;
		    if(result[j] != ptr->fid) {
		      result[i-1]=ptr->fid;
		      ++i;
		      result=realloc(result, i*sizeof(uint32_t));
		      result[i-1]=0;
		    }
		  }
		}
		free(search_in);
	}
	free(ival);
	return result;
}
