/*
 * libkarma/properties.h
 * 
 * Copyright (c) Frank Zschockelt <libkarma@freakysoft.de> 2004
 * 
 * You may distribute and modify this program under the terms of 
 * the GNU GPL, version 2 or later.
 * 
 */
#include <iconv.h>
#include <inttypes.h>

#include "hash.h"

#ifdef __cplusplus
extern "C" {
#endif

struct list{
	uint32_t fid;
	HASH * element;
	struct list * next;
};

struct klist{
	struct list * l;
	uint32_t count;
	uint32_t highest_fid;
	iconv_t in;
	iconv_t out;
	char * charset;
};

void     init_properties     (char * codeset);
void     destroy_properties  (void);

char *   current_codeset     (void);
void     change_codeset      (char * codeset);
char *   convert2utf8        (char * string);
char *   convertFromUtf8     (char * string);

uint32_t import_properties   (char * properties);
char *   export_property     (uint32_t id);

uint32_t getnextfid          (void);

uint32_t new_property        (void);
int      delete_property     (uint32_t id);
char *   get_property        (uint32_t id, char * key);
char *   get_propertyUtf8    (uint32_t id, char * key);
char *   get_property_hash   (HASH * p, char * key);
int      change_property     (uint32_t id, char * property, char * data);
void     change_property_hash(HASH * p, char * property, char * data);
void 	 change_property_hash_raw(HASH * p, char * property, char * data);

HASH *   idsearch            (uint32_t fid);

int      load_properties     (void);
int      save_properties     (void);
void     inc_devicegeneration(void);
int      cache_obsolete      (char *device_settings);
int      db_changed          (char *device_settings);

/* uint32_t * search(uint32_t * search_in, char * key, char * data); */
uint32_t *andOrSearch(int andOr, uint32_t *search_in, char *key, char *data);

#ifdef __cplusplus
}
#endif  /* C++ */

