/*
 * libkarma/util.c
 *
 * Copyright (c) Frank Zschockelt <libkarma@freakysoft.de> 2004
 *
 * char *simple_itoa was borrowed from boa [http://www.boa.org]
 *
 * You may distribute and modify this program under the terms of 
 * the GNU GPL, version 2 or later.
 * 
 */
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <errno.h>

#include "errors.h"
#include "util.h"


/*copied from the boa webserver*/
char *simple_itoa(unsigned int i)
{
	/* 21 digits plus null terminator, good for 64-bit or smaller ints
	 * for bigger ints, use a bigger buffer!
	 *
	 * 4294967295 is, incidentally, MAX_UINT (on 32bit systems at this time)
	 * and is 10 bytes long
	 */
	static char local[22];
	char *p = &local[21];
	*p-- = '\0';
	do{
		*p-- = '0' + i%10;
		i/=10;
	}while(i>0);
	return p+1;
}

/* Recursively create directories in a path. The last element is assumed to
 *    be a filename, and simply ignored. Code based on create_bittorrent_path() */
int create_path(const char *pathname)
{
	int pos, ret;
	char *path;
	unsigned char separator;
	
	path=strdup(pathname);
	if (!*path) return -1;
	
	for (pos = 1; path[pos]; pos++) {
		separator = path[pos];
		if (separator != '/')
			continue;
		path[pos] = 0;
		ret = mkdir(path, 0700); /*S_IREAD | S_IWRITE | S_IEXEC);*/
		path[pos] = separator;
		if (ret < 0 && errno != EEXIST){
			free(path);
			lkerrSet(E_MKDIR);
			return -1;
		}
	}
	free(path);
	return 0;
}
